/*
 *
 *              Reverse MX Filter for Postfix 2
 *                    reject_bad_rmx
 *           (c) 2004 Elita rozanski@sergiusz.com
 *                    v. 2004.02.05
 *
 */

/*
MXfilter is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.
*/

#include <netinet/in.h>
#include <resolv.h>
#include <sys/socket.h>
#include <arpa/nameser.h>
#include <stdio.h>
#include <netdb.h>
#include <syslog.h>

#include "utils.c"
#include "mxget.c"
#include "fecyk.c"

// only one recipient address
// #define TESTER_ONLY

/*
The variable h_errno can have the following values:
HOST_NOT_FOUND
The specified host is unknown.
NO_ADDRESS or NO_DATA
The requested name is valid but does not have an IP address.
NO_RECOVERY
A non-recoverable name server error occurred.
TRY_AGAIN
A temporary error occurred on an authoritative name
server.  Try again later.
*/


/* 
   test access list for mxfilter
   return 0 for skip
   return 1 for check
   return < 0 for error
*/

int mxfilter_acl(char *recipient, char *result, char *reason) {
FILE *p;
char buf[1024],*k;
char r[1024];
int i;

result[0]=0;
reason[0]=0;

if(strlen(recipient)==0)return -1; //recipient empty

snprintf(r,sizeof(r),"%s",recipient);

for(i=0;i<strlen(r);i++)
    r[i]=tolower(r[i]);

if(!(p=fopen("/etc/postfix/mxfilter.acl","r"))) return -2; //acl error

while(fgets(buf,sizeof(buf)-1,p)) {
    k=(char *)index(buf,10);
    if(k)k[0]=0;
    k=(char *)index(buf,13);
    if(k)k[0]=0;
    
    if(strlen(buf)>0) { // empty line ignore
	if(buf[0]!='#') { // remark line ignore
	    k=(char *)index(buf,9);
	    if(k) { //bad line ignore
		k++[0]=0;
// printf("* %s %s %s\n",recipient,buf,k);
		if(strstr(r,buf) || strcmp(buf,"*")==0) { // ok
		    if(strncmp(k,"check",5)==0) {
			k=(char *)index(k,9);
			if(k) {
			    k++;
			    sprintf(result,"%s",k);
			    k=(char *)index(result,9);
			    if(k) {
				k++[0]=0;
				sprintf(reason,"%s",k);
				}
			    }
// printf("** %s %s\n",result,reason);
			return 1;
			}
		    if(strcmp(k,"skip")==0)
			return 0;
		    // other action ignore
		    }
		}
	    }    
	}
    }
fclose(p);
return 0; //default skip
}

int check_mx(char *sender,char *client_address,int tolerance) {
struct hostent *hh;
int i,f;
char *domain;
char relay[1024];
char user[1024];
char fquery[1024];

if (!index(sender,'@')) return 1; //invalid email address
domain=index(sender,'@')+1;
if (strlen(domain)<3) return 1; //domain too short

snprintf(user,sizeof(user),"%s",sender);
index(user,'@')[0]=0;

if (strlen(user)==0)return 1; //username too short

/* simple fecyk query (ehlo ignored)*/

// _smtp-client.domain.tld first
snprintf(fquery,sizeof(fquery),"_smtp-client.%s",domain);
f=find_dmp(fquery);

if(f==0) { //domain has dmp records
    char rev_ca[100];

// rev_client_address.in-addr._smtp-client.domain.tld
    
    rev_addr(client_address,rev_ca);
//    printf("fecyk %s, %s",client_address,rev_ca);
    snprintf(fquery,sizeof(fquery),"%s.in-addr._smtp-client.%s",rev_ca,domain);
    f=find_dmp(fquery);
    if(f==1) {
	return 0; //ok dunno
	}
    if(f==2 || f==(-2) || f==0) {
	return 2; //adm reject or too many answers or empty (acl action (550))
	}
    if(f==(-1)) {
	return (-1); //dns error 450
	}
    };
//else try RMXF other tests

/* a 'user.relay' */
snprintf(relay,sizeof(relay),"%s.relay.%s",user,domain);
hh=gethostbyname(relay);
if(!hh) {
    if(h_errno==TRY_AGAIN || h_errno==NO_RECOVERY)
	return (-1); //dns error
    }
else {
for(i=0;i<256;i++) {
    if(hh->h_addr_list[i]) {
	char odp[256];
	snprintf(odp,sizeof(odp),"%d.%d.%d.%d",(unsigned char)hh->h_addr_list[i][0],
	    (unsigned char)hh->h_addr_list[i][1],
	    (unsigned char)hh->h_addr_list[i][2],
	    (tolerance==1)?0:(unsigned char)hh->h_addr_list[i][3]);
	if (strncmp(odp,client_address,strlen(odp)-tolerance)==0) {
	    return 0; //ok
	    }
	if (strncmp(odp,"127.0.0",7)==0) {
	    return 2; //admin reject
	    }
	}
    else {
	i=256;
	}
    }
    }

/* a 'relay' */
snprintf(relay,sizeof(relay),"relay.%s",domain);
hh=gethostbyname(relay);
if(!hh) {
    if(h_errno==TRY_AGAIN || h_errno==NO_RECOVERY)
	return (-1); //dns error
    }
else {
for(i=0;i<256;i++) {
    if(hh->h_addr_list[i]) {
	char odp[256];
	snprintf(odp,sizeof(odp),"%d.%d.%d.%d",(unsigned char)hh->h_addr_list[i][0],
	    (unsigned char)hh->h_addr_list[i][1],
	    (unsigned char)hh->h_addr_list[i][2],
	    (tolerance==1)?0:(unsigned char)hh->h_addr_list[i][3]);
	if (strncmp(odp,client_address,strlen(odp)-tolerance)==0) {
	    return 0;
	    }
	if (strncmp(odp,"127.0.0",7)==0) {
	    return 2; //admin reject
	    }
	}
    else {
	i=256;
	}
    }
    }

/* a */
hh=gethostbyname(domain);
if(!hh) {
    if(h_errno==TRY_AGAIN || h_errno==NO_RECOVERY)
	return (-1); //dns error
    }
else {
for(i=0;i<256;i++) {
    if(hh->h_addr_list[i]) {
	char odp[256];
	snprintf(odp,sizeof(odp),"%d.%d.%d.%d",(unsigned char)hh->h_addr_list[i][0],
	    (unsigned char)hh->h_addr_list[i][1],
	    (unsigned char)hh->h_addr_list[i][2],
	    (tolerance==1)?0:(unsigned char)hh->h_addr_list[i][3]);
	if (strncmp(odp,client_address,strlen(odp)-tolerance)==0) {
	    return 0;
	    }
	if (strncmp(odp,"127.0.0",7)==0) {
	    return 2; //admin reject
	    }
	}
    else {
	i=256;
	}
    }
    }

/* mx */
i=find_mailhost(domain,client_address,tolerance);

return i;
}

int main (void) {
#ifdef TESTER_ONLY
char tester[]="rozanski@sergiusz.com";
#endif
int log=1;

char buf[1001];
char *k;

char sender[257];
char client_address[257];
char recipient[257];

// int i=0;

memset(&sender,0,sizeof(sender));
memset(&client_address,0,sizeof(client_address));
memset(&recipient,0,sizeof(recipient));

/* input
sender=ktos@comm.pl
client_address=213.25.233.144
recipient=rozanski@sergiusz.com
*/

while((fgets(buf,1000,stdin))) {
    k=(char *)index(buf,10);
    if(k)k[0]=0;
    k=(char *)index(buf,13);
    if(k)k[0]=0;

//    printf("%d %s\n",i++,buf);

    if(strncmp(buf,"sender=",7)==0) {
	snprintf(sender,sizeof(sender),"%s",index(buf,'=')+1);
	}
    if(strncmp(buf,"client_address=",15)==0) {
	snprintf(client_address,sizeof(client_address),"%s",index(buf,'=')+1);
	}
    if(strncmp(buf,"recipient=",10)==0) {
	snprintf(recipient,sizeof(recipient),"%s",index(buf,'=')+1);
	}
    if(strlen(buf)==0) {
	int result=0;
	char exitcode[30]="";
	char reason[1024]="";
#ifndef TESTER_ONLY
	int acl;
#endif
	
	k=(char *)index(sender,'@');
	if(!k) {
	    if(log) {
		openlog("postfix/mxfilter", 0, LOG_MAIL);
		syslog(0,"%s %s %s %s\n",client_address,sender,recipient,"ERROR: bad sender");
		closelog();
		}
	    printf("action=550 Sender address rejected\n\n"); //not @ in sender?, bug in postfix?
	    return 0;
	    }

#ifdef TESTER_ONLY
	if(strcmp(recipient,tester)==0) {
	    result=check_mx(sender,client_address,1);
	    }
#else
	acl=mxfilter_acl(recipient,exitcode,reason);
	if(acl<0) { //acl file error
	    if(log) {
		openlog("postfix/mxfilter", 0, LOG_MAIL);
		syslog(0,"%s %s %s %s\n",client_address,sender,recipient,"ERROR: access list");
		closelog();
		}
	    printf("action=450 Email system error: mxfilter.acl problem. http://rmxf.comm.pl/\n\n");
	    return 0;
	    }
	if(acl)
	    result=check_mx(sender,client_address,1);
#endif

	if(result==0) {
	    printf("action=dunno\n\n"); //ok MX or A record
	    }
	if(result==1) {
	    char lres[1024];
	    if(strcmp(exitcode,"HMARK")==0) {
		printf("action=prepend X-mxfilter: rmx-false\t\ndunno\n\n");
		}
	    else {
		snprintf(lres,sizeof(lres),"Host %s not MX or A for domain %s. Check dns or use true relay host. http://rmxf.comm.pl/",
	    	    client_address,
	    	    index(sender,'@')+1);
		printf("action=%s %s\n\n",
		    (strlen(exitcode)==0)?"550":exitcode,
		    (strlen(reason)==0)?lres:reason);
		}
	    }
	if(result==2) {
	    printf("action=550 Address %s rejected by domain administrator. http://rmxf.comm.pl/\n\n",
	        sender);
	    }
	if(result==(-1)) {
	    printf("action=450 DNS Error try again later. http://rmxf.comm.pl/\n\n");
	    }

	if(log) {
	    openlog("postfix/mxfilter", 0, LOG_MAIL);
	    syslog(0,"%s %s %s %s\n",
		client_address,
		sender,
		recipient,
		(result==0)?"OK":(result==1)?"REJECT":(result==2)?"ADMREJECT":"DNSERROR");
	    closelog();
	    }

	return 0;
	}
    }

// strange, postfix false?
if(log) {
    openlog("postfix/mxfilter", 0, LOG_MAIL);
    syslog(0,"%s %s %s %s\n",client_address,sender,recipient,"ERROR: stdin empty");
    closelog();
    }
printf("action=dunno\n\n");
return 0;
}
